/*
Vertical Pairing Bracket and Bout Sheet Maker
Copyright (C) 2010 name of author

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/
import groovy.swing.SwingBuilder

import java.awt.Font
import java.io.File
import java.io.FileReader
import java.io.BufferedOutputStream
import java.io.BufferedReader
import java.io.PrintStream

import java.util.Enumeration
import javax.swing.AbstractButton

import javax.swing.JFileChooser
import javax.swing.JOptionPane
import javax.swing.WindowConstants
//import javax.swing.filechooser.FileNameExtensionFilter;

class BoutMaker
{
	static final int MAX_ENTRY = 10
	def app = new PDFMaker()
	SwingBuilder swing = new SwingBuilder()
	CompetitorInfo infoBlock = new CompetitorInfo( )
	
	def clearAction = swing.action( name: 'Clear Names', closure: this.&clearNames )
	def makeAction = swing.action( name: 'Make PDF', closure: this.&make )
	def openAction = swing.action( name: 'Open...', mnemonic: 'O', accelerator: 'ctrl O',
		closure: this.&openInfoFile )
	def saveAction = swing.action( name: 'Save...', mnemonic: 'S', accelerator: 'ctrl S',
		closure: this.&saveInfoFile )
	def swapAction = swing.action( name: 'Swap', closure: this.&swapNames )
	def exitAction = swing.action( name: 'Exit', mnemonic: 'x', accelerator: 'ctrl Q',
		closure: this.&goodbye )
	def setNumberingAction = swing.action(name: "numbering", closure: this.&setNumbering )
	
	def numberStart = 1;
	def numberDelta = 1;
	
	void clearNames( evt )
	{
		1.upto(10) { n ->
			swing.getProperty("name${n}").text = ""
			swing.getProperty("team${n}").text = ""
		}
	}

	void setNumbering( evt )
	{
		String str = swing.numbering.selectedItem;
		if (str.equals("sequential"))
		{
			numberStart = 1;
			numberDelta = 1;
		}
		else if (str.startsWith("odd"))
		{
			numberStart = 1;
			numberDelta = 2;
		}
		else if (str.startsWith("even"))
		{
			numberStart = 2;
			numberDelta = 2;
		}
		renumber();
	}
	
	void renumber()
	{
		1.upto(10) { n ->
			swing.getProperty("c${n}").text = 
				(String) "${(n-1) * numberDelta + numberStart}"
		}
	}

	void swapNames(evt)
	{
		int fromLine
		int toLine
		String fromName
		String fromTeam
		
		String nameFrom
		String teamFrom
		String nameTo
		String teamTo
		
		fromLine = Integer.parseInt(swing.getProperty("swapFrom").text);
		toLine = Integer.parseInt(swing.getProperty("swapTo").text);
		
		/* take account of the numbering scheme */
		fromLine = (fromLine - numberStart) / numberDelta + 1
		toLine = (toLine - numberStart) / numberDelta + 1
		if (fromLine > 0 && fromLine <= 10)
		{
			nameFrom = "name${fromLine}"
			teamFrom = "team${fromLine}"
			nameTo = "name${toLine}"
			teamTo = "team${toLine}"
			fromName = swing.getProperty(nameFrom).text;
			fromTeam = swing.getProperty(teamFrom).text;
			swing.getProperty(nameFrom).text = swing.getProperty(nameTo).text;
			swing.getProperty(teamFrom).text = swing.getProperty(teamTo).text;
			swing.getProperty(nameTo).text = fromName;
			swing.getProperty(teamTo).text = fromTeam;
		}
	}
	
	CompetitorInfo collectInfo( )
	{
		CompetitorInfo infoBlock = new CompetitorInfo( )
		String name
		String team

		infoBlock.tournament = swing.tournament.text;
		infoBlock.style = swing.style.selectedItem
		infoBlock.ageGroup = swing.ageGroup.selectedItem
		infoBlock.weight = swing.weight.text
		infoBlock.ageStyle = (infoBlock.style.equals("GR")) ?
			"${infoBlock.ageGroup} GRECO" :
			infoBlock.ageGroup
		
		for (i in 1 .. MAX_ENTRY)
		{
			name = swing.getProperty( "name${i}" ).text
			team = swing.getProperty( "team${i}" ).text 
			if (!(name =~ /^\s*$/))
			{
				infoBlock.nameList << name
				infoBlock.teamList << team
			}
		}
		infoBlock.makeCopy = swing.getProperty("makeCopy").selected
		infoBlock.numberingStyle = swing.numbering.selectedItem
		return infoBlock
	}

	void make( event )
	{
		infoBlock = collectInfo( )
	
		if (infoBlock.nameList.size > 1)
		{
			app.constructPDF( infoBlock )
		}
		else
		{
			JOptionPane.showMessageDialog(null, "You must have at least two competitors.",
				"Error", JOptionPane.ERROR_MESSAGE) 
			// JOptionPane.showMessageDialog(null, "You must have at least two competitors.")
		}
	}
	
	void openInfoFile( evt )
	{
		BufferedReader inFile
		File f
		int n
		String name
		String team

		f = getFile( "", false )
		if (f != null)
		{
			inFile = new BufferedReader( new FileReader( f ) )
			infoBlock.tournament = inFile.readLine()
			swing.tournament.text = infoBlock.tournament
			
			infoBlock.style = inFile.readLine( )
			swing.style.setSelectedItem( infoBlock.style )
			
			infoBlock.ageGroup = inFile.readLine( )
			swing.ageGroup.setSelectedItem( infoBlock.ageGroup )
			
			infoBlock.weight = inFile.readLine( )
			swing.weight.text = infoBlock.weight

			infoBlock.ageStyle = inFile.readLine( )
			
			infoBlock.numberingStyle = inFile.readLine( )
			swing.numbering.setSelectedItem( infoBlock.numberingStyle )
			renumber()
			
			infoBlock.nameList = []
			infoBlock.teamList = []
			n = Integer.parseInt( inFile.readLine( ) )
			for (i in 0 ..< n)
			{
				name = inFile.readLine()
				team = inFile.readLine()
				infoBlock.nameList << name
				infoBlock.teamList << team
				if (i > 0)
				{
					swing.getProperty( "name${i}").text = name
					swing.getProperty( "team${i}" ).text = team
				}
			}
			if (n < MAX_ENTRY)
			{
				for (i in n .. MAX_ENTRY)
				{
					swing.getProperty( "name${i}").text = ""
					swing.getProperty( "team${i}").text = ""
				}
			}	
			inFile.close( )
		}
	}
	
	void saveInfoFile( evt )
	{
		PrintStream outFile
		File f
		String fileName
		infoBlock = collectInfo( )
		if (infoBlock.nameList.size() > 1)
		{
			fileName = swing.ageGroup.selectedItem + "~" +
				swing.weight.text + "~" + swing.style.selectedItem
			fileName = fileName.replaceAll(/[^A-Za-z_0-9.]/) { "~" }
			fileName = fileName + ".dat"
			f = getFile(fileName.toLowerCase(), true)
		}
		if (f != null)
		{
			outFile = new PrintStream( f )
			outFile.println( infoBlock.tournament )
			outFile.println( infoBlock.style )
			outFile.println( infoBlock.ageGroup )
			outFile.println( infoBlock.weight )
			outFile.println( infoBlock.ageStyle )
			outFile.println( infoBlock.numberingStyle )
			outFile.println( infoBlock.nameList.size() )
			for (i in 0 ..< infoBlock.nameList.size())
			{
				outFile.println( infoBlock.nameList[i] )
				outFile.println( infoBlock.teamList[i] )
			}
			outFile.close()
		}
		
	}

	File getFile( fileName, saving )
	{
		JFileChooser chooser = new JFileChooser( "." );
		/*
		FileNameExtensionFilter filter = new FileNameExtensionFilter(
        	"Saved Data Files", "dat")
		*/
		DataFileFilter filter = new DataFileFilter();
		chooser.setFileFilter(filter)
		chooser.setSelectedFile(new File(fileName))
		File result = null
		
		int returnVal
		if (saving)
		{
			returnVal = chooser.showSaveDialog( null );
		}
		else
		{
			returnVal = chooser.showOpenDialog( null );
		}
		if (returnVal == JFileChooser.APPROVE_OPTION)
		{
			result = chooser.getSelectedFile();
		}
		return result
	}

	void goodbye( evt )
	{
		System.exit( 0 )
	}

	def bigFont = new Font( "Helvetica", Font.PLAIN, 18)
	
	void doIt( )
	{
		def frame = swing.frame(title: 'Boutsheet Maker',
			defaultCloseOperation:WindowConstants.DISPOSE_ON_CLOSE) {
			menuBar {
				menu('File') {
					menuItem( action: openAction )
					menuItem( action: saveAction )
					menuItem( action: exitAction )
				}
			}
			panel {
				vbox {
					hbox {
						label( 'Tournament' )
						textField(id: 'tournament', columns: 25)
					}
					rigidArea( )
					hbox {
						label('Age Group' )
						comboBox( id: 'ageGroup', items: ['Rookie', 'Bantam', 'Intermediate', 'Novice',
							'Schoolboy/girl', 'Cadet', 'Junior', 'Open'], selectedIndex: 1 )
						rigidArea( )
						label('Weight')
						textField(id: 'weight', columns: 7)
						rigidArea( )
						label('Style')
						comboBox( id: 'style', items: ['FS', 'GR'] )
					}
					rigidArea( )
					tableLayout {
						tr {
							td { label "" }
							td(align: 'CENTER') { label 'Name' }
							td(align: 'CENTER') { label 'Team' }
						}
						1.upto(10) { n ->
							tr {
								td { label(id: "c${n}", text: (String) "${(n-1) * numberDelta + numberStart}" )}
								td { textField( font: bigFont, id: "name${n}", columns: 20 ) }
								td { textField( font: bigFont, id: "team${n}", columns: 20 ) }
							} 
						}
						tr {
							td(colspan:3) {
								hbox {
								checkBox(id: 'makeCopy', selected: true)
								label('Make copy of first page')
								rigidArea()
								button(action: makeAction)
								rigidArea()
								button(action: clearAction)
								rigidArea()
								label('Number competitors: ')
								comboBox( id: 'numbering', items: ['sequential', 'odd #s', 'even #s'],
									action: setNumberingAction)
								}
							}
						}
					}
					rigidArea()
					hbox {
						label('Swap Lines')
						rigidArea()
						textField(id: 'swapFrom', columns: 3)
						rigidArea()
						label('and')
						rigidArea()
						textField(id: 'swapTo', columns: 3)
						rigidArea()
						button(action: swapAction)
					}
				}
			}
		}
		
		frame.pack( )
		frame.show( )
	}
	
	public static void main(String [] args)
	{
		def app = new BoutMaker();
		app.doIt()
	}

}
