import java.io.File
import java.io.FileReader
import java.io.BufferedOutputStream
import java.io.BufferedReader
import java.io.PrintStream
import java.text.SimpleDateFormat
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import javax.swing.JOptionPane
import javax.swing.JFileChooser

//JAXP
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.Source;
import javax.xml.transform.Result;
import javax.xml.transform.stream.StreamSource;
import javax.xml.transform.sax.SAXResult;

// FOP
import org.apache.fop.apps.FOUserAgent;
import org.apache.fop.apps.Fop;
import org.apache.fop.apps.FOPException;
import org.apache.fop.apps.FopFactory;
import org.apache.fop.apps.FormattingResults;
import org.apache.fop.apps.MimeConstants;
import org.apache.fop.apps.PageSequenceResults;

class PDFMaker
{
	static final int MAX_ENTRY = 10
	
	// pairing sheet is measured in cm
	static final double PAIR_BOUT_LEFT = 7.25
	static final double PAIR_HEADING_WIDTH = 1.25
	static final double PAIR_HEADING_HEIGHT = 0.4
	static final double PAIR_BOUT_HEIGHT = 1.5
	static final double PAIR_BOUT_WIDTH = 1.25
	static final double PAIR_BOUT_LOW_Y = 3.85
	static final double PAIR_ROUNDS = 7
	static final double PAIR_NUMBER_LEFT = 1.0
	static final double PAIR_NUMBER_TOP = 4.7
	static final double PAIR_NAME_LEFT = 2.4
	static final double PAIR_NAME_TOP = 4.6
	static final double PAIR_TEAM_TOP = PAIR_NAME_TOP + 0.75
	static final double PAIR_TEAM_LEFT = 2.4
	
	static final double SEMI_OFFSET = 0.3	/* move semicircle left a bit */
	
	FopFactory fopFactory = FopFactory.newInstance();
	
	def pairings = [
		[0],
		[0],
		[[1,2],[1,2],[1,2]],
		[[1,2,3,0], [3,1,2,0], [2,3,1,0]],
		[[1,2,3,4], [1,3,2,4]],
		[[1,2,3,4,5,0], [5,1,2,3,4,0]],
		[[1,2,3,4,5,6], [1,3,2,5,4,6]],
		[[1,2,3,4,5,6,7,0], [7,1,2,3,4,5,6,0]],
		[[1,2,3,4,5,6,7,8], [1,3,2,4,5,7,6,8]],
		[[1,2,3,4,5,6,7,8,9,0], [9,1,2,3,4,5,6,7,8,0]],
		[[1,2,3,4,5,6,7,8,9,10], [1,3,2,4,5,7,6,9,8,10]]
	]
		
	def preBoutSVG = []
	def postBoutSVG = []
	
	String tempFilename = ""
	String pdfFilename	
	File tempFile
	File pdfFile
	PrintStream outFile
	BufferedReader inFile
	BufferedOutputStream pdfStream
	
	String date
	
	CompetitorInfo infoBlock
	
	int numberStart;
	int numberDelta;
	
	PDFMaker()
	{
		String data

		inFile = new BufferedReader(new FileReader("lib/boutpage.fo"))
		while (((data = inFile.readLine()) != null) && (! (data =~ /<!--/)))
		{
			preBoutSVG << data
		}
		
		while ((data = inFile.readLine()) != null)
		{
			postBoutSVG << data
		}
		
		inFile.close()	
	}

	void constructPDF(infoBlock)
	{
		String data
		date = getDate()
		this.infoBlock = infoBlock;
		
		setNumbering()

		/*	Create the .fo file that will be used as input to FOP */
		tempFile = File.createTempFile("bout", ".fo")
		// tempFile = new File("testor.fo") /* debug */
		tempFilename = tempFile.getAbsolutePath()
		
		inFile = new BufferedReader(new FileReader("lib/template.fo"))
		outFile = new PrintStream(tempFile)
		while ((data = inFile.readLine()) != null)
		{
			if (data =~ /<!--\*\*NAMES\*\*-->/)
			{
				insertNames()
				if (infoBlock.makeCopy)
				{
					insertNames()
				}
			}
			else if (data =~ /<!--\*\*BOUTS\*\*-->/)
			{
				insertBouts()
			}
			else
			{
				outFile.println(data)
			}
		}
		inFile.close()
		outFile.close()
		
		pdfFilename = infoBlock.ageGroup + "~" + infoBlock.weight + "~" + infoBlock.style
		pdfFilename = pdfFilename.replaceAll(/[^A-Za-z_0-9.]/) { "~" }
		pdfFilename += ".pdf"
		pdfFilename = pdfFilename.toLowerCase();
		
		if (true)
		{
			pdfStream = new BufferedOutputStream(new FileOutputStream(new File(pdfFilename)))
			try {
				// Step 3: Construct fop with desired output format
				Fop fop = fopFactory.newFop(MimeConstants.MIME_PDF, pdfStream)
				
				// Step 4: Setup JAXP using identity transformer
				TransformerFactory factory = TransformerFactory.newInstance()
				Transformer transformer = factory.newTransformer() // identity transformer
						
				// Step 5: Setup input and output for XSLT transformation 
				// Setup input stream
				Source src = new StreamSource(new File(tempFilename))
				
				// Resulting SAX events (the generated FO) must be piped through to FOP
				Result res = new SAXResult(fop.getDefaultHandler())
							
				// Step 6: Start XSLT transformation and FOP processing
				transformer.transform(src, res)
				
			} finally {
				//Clean-up
				pdfStream.close();
				JOptionPane.showMessageDialog(null, "File ${pdfFilename} created.");
				tempFile.delete();
			}
		}
	}
	
	File getFile(fileName)
	{
		JFileChooser chooser = new JFileChooser(".");
		File result = null
		// Note: source for ExampleFileFilter can be found in FileChooserDemo,
		// under the demo/jfc directory in the JDK.
		
		int returnVal = chooser.showSaveDialog(null);
		if (returnVal == JFileChooser.APPROVE_OPTION)
		{
			result = chooser.getSelectedFile();
		}
		return result
	}

	void insertNames()
	{
		double xPos
		double yPos
		int nCompetitors = infoBlock.nameList.size() - 1
		def pairList = pairings[nCompetitors]
		def boutList = []
		def numberList = []
		int bout
		int round
		
		outFile.println("""
<fo:page-sequence master-reference=\"chart\">
<fo:flow flow-name=\"xsl-region-body\" font-family=\"Times\">""");

		/* Headings for the pairing sheet
			Yeah, there are a lot of manifest constants here. Life is rough. */
		makeTextCm(5, 1.4, 12, infoBlock.tournament)
		makeTextCm(2.6, 3.1, 12, infoBlock.ageStyle)
		makeTextCm(10.2, 3.1, 12, infoBlock.weight)
		makeTextCm(15.3, 3.1, 12, date)

		for (i in 1 .. nCompetitors)
		{
			makeTextCm(PAIR_NUMBER_LEFT,
				PAIR_NUMBER_TOP + (i-1) * PAIR_BOUT_HEIGHT, 18, generateSequenceNumber(i))
	
			makeTextCm(PAIR_NAME_LEFT,
				PAIR_NAME_TOP + (i-1) * PAIR_BOUT_HEIGHT, 10, infoBlock.nameList[i])
	
			makeTextCm(PAIR_TEAM_LEFT,
				PAIR_TEAM_TOP + (i-1) * PAIR_BOUT_HEIGHT, 10, infoBlock.teamList[i])
		}
		
		/*
			Now the bout numbers
		*/
		round = 0; // so that x-offset is easy to calculate
		xPos = 7.32
		if (nCompetitors < 4)
		{
			xPos += 7 * 1.25
			for (i in 0 ..< nCompetitors)
			{
				semicircle(xPos - SEMI_OFFSET, 4.62 + i * PAIR_BOUT_HEIGHT)
			}
		}
	
		crossout(nCompetitors)
	
		pairList.each
		{	item ->
			yPos = 4.62
			boutList = item
			numberList = []
			
			bout = 0
			while(bout < boutList.size())
			{
				numberList[boutList[bout]] = boutList[bout + 1]
				numberList[boutList[bout + 1]] = boutList[bout]
				bout += 2
			}
			
			for (i in 1 ..< numberList.size())
			{
				outFile.print("<fo:block-container ")
				outFile.println("position=\"absolute\" left=\"${xPos}cm\" top=\"${yPos}cm\">")
				outFile.println("<fo:block font-size=\"8pt\">")
				outFile.print((numberList[i] == 0) ? "Bye" : generateSequenceNumber(numberList[i]))
				outFile.println("</fo:block>")
				outFile.println("</fo:block-container>")
				yPos += PAIR_BOUT_HEIGHT
			}
			xPos += PAIR_BOUT_WIDTH
		}
		outFile.println("""
	<fo:block-container position=\"absolute\">
		<fo:block>
			<fo:external-graphic width=\"8.5in\" height=\"11in\"
				position=\"absolute\" left=\"0cm\" top=\"0cm\"
				src=\"lib/wallchart.svg\"/>
		</fo:block>
	</fo:block-container>
		
</fo:flow>
</fo:page-sequence>""");
	}

	void insertBouts()
	{
		int nCompetitors = infoBlock.nameList.size() - 1
		int round = 1
		String roundPrefix = (nCompetitors < 4) ? "RR" : ""
		def pairList = pairings[nCompetitors]
		def boutList = []
		int bout
		double xOffset
		double yOffset
		double scoreOffset
		final double SVG_X_OFFSET = 5.25
		final double SCORE_OFFSET = 2.5
		int competitor

		pairList.each
		{
			item ->
			competitor = 0
			boutList = item
			bout = 0
			while (bout < boutList.size() / 2)
			{
				if (bout % 2 == 0)
				{
					preBoutSVG.each
					{
						outFile.println(it)
					}
				}
				xOffset = (bout % 2 == 0) ? 0 : SVG_X_OFFSET
				
				makeTextInches(xOffset + 0.1, 0.5, 18, infoBlock.ageStyle)
				makeTextInches(xOffset + 0.1, 0.75, 18, infoBlock.weight)
				makeTextInches(xOffset + 0.7, 1.55, 12, infoBlock.weight)
				makeTextInches(xOffset + 2.7, 1.55, 12, "${roundPrefix}${round}")
				makeTextInches(xOffset + 1.9, 1.55, 12, infoBlock.style)
				
				for (j in 0..1)
				{
					scoreOffset = (j == 0) ? 0 : SCORE_OFFSET
					if (boutList[bout * 2 + j] != 0)
					{
						makeTextInches (xOffset + scoreOffset + 0.1, 2.25,
							18, generateSequenceNumber(boutList[bout * 2 + j]))
					}
					makeTextInches(xOffset + scoreOffset + 0.4, 2.15, 12,
						infoBlock.nameList[boutList[bout * 2 + j]])
					makeTextInches(xOffset + scoreOffset + 0.4, 2.4, 12,
						infoBlock.teamList[boutList[bout * 2 + j]])
				}
				if (bout % 2 == 1)
				{
					postBoutSVG.each
					{
						outFile.println(it)
					}
				}
				bout++
			}
			if (bout % 2 == 1)
			{
				postBoutSVG.each
				{
					outFile.println(it)
				}
			}
			round++
		}
	}

	/*
		Draw a semicircle for round-robin matches at the given x, y coordinate
	*/
	void semicircle(double x, double y)
	{
		outFile.println("""
<fo:block-container position=\"absolute\" left=\"${x}cm\" top=\"${y}cm\">
<fo:block>
<fo:instream-foreign-object width=\"0.2cm\" height=\"1.5cm\">
<svg xmlns=\"http://www.w3.org/2000/svg\" width=\"0.2cm\" height=\"1.25cm\"
	viewBox=\"0 0 0.2 1.5\">
		<path style=\"fill:red; stroke: none;\"
			d=\"M 0.2 0 A 0.2 0.625 0 1 0 0.2 1.25 L 0.2 0\" />
</svg>
</fo:instream-foreign-object>
</fo:block>
</fo:block-container>""")
	}
	/*
		In the next three functions, message is untyped so that
		duck typing can print either a double or a String.
	*/
	void makeTextInches(double x, double y, double fontSize, message)
	{
		makeTextUnits(x, y, fontSize, message, "in")
	}
	
	void makeTextCm(double x, double y, double fontSize, message)
	{
		makeTextUnits(x, y, fontSize, message, "cm")
	
	}
	
	void makeTextUnits(double x, double y, double fontSize, message, String units)
	{
		
		if (message != "")
		{
			outFile.print("<fo:block-container position=\"absolute\" ")
			outFile.println("left=\"${x}${units}\" top=\"${y}${units}\">")
			outFile.println("<fo:block font-size=\"${fontSize}pt\">${message}</fo:block>")
			outFile.println("</fo:block-container>")
		}
	}
	
	
	/*
		Cross out the "we wrote bout sheets" area
	*/
	void crossout(int nPeople)
	{
		int limit = (nPeople < 4) ? 3 : 2
		double x = PAIR_BOUT_LEFT + ((nPeople < 4) ? PAIR_ROUNDS * PAIR_HEADING_WIDTH : 0)
		double y = 3.85
		
		outFile.print("""<fo:block-container position=\"absolute\" left=\"${x}cm\" top=\"${y}cm\">
	<fo:block>
	<fo:instream-foreign-object
		width=\"${3*PAIR_HEADING_WIDTH}cm\" height=\"${PAIR_HEADING_HEIGHT}cm\">
	<svg xmlns=\"http://www.w3.org/2000/svg\"
		width=\"${3*PAIR_HEADING_WIDTH}cm\" height=\"${PAIR_HEADING_HEIGHT}cm\"
		viewBox=\"0 0 ${3*PAIR_HEADING_WIDTH} ${PAIR_HEADING_HEIGHT}\">
			<path style=\"fill:none; stroke: black; stroke-width: 0.0330;\" """)
	
		outFile.print("d=\"")
		for (i in 0 ..< limit)
		{
			outFile.print("M${i*PAIR_BOUT_WIDTH} ${PAIR_HEADING_HEIGHT} L ${(i+1)*PAIR_BOUT_WIDTH} 0 ")
		}
		outFile.println("\"/>")
		outFile.println("""</svg>
	</fo:instream-foreign-object>
	</fo:block>
	</fo:block-container>""")
	}

	int generateSequenceNumber(int n)
	{
		return ((n - 1) * numberDelta + numberStart)
	}

	void setNumbering( )
	{
		if (infoBlock.numberingStyle.equals("sequential"))
		{
			numberStart = 1;
			numberDelta = 1;
		}
		else if (infoBlock.numberingStyle.startsWith("odd"))
		{
			numberStart = 1;
			numberDelta = 2;
		}
		else if (infoBlock.numberingStyle.startsWith("even"))
		{
			numberStart = 2;
			numberDelta = 2;
		}
	}
	
	String getDate() {
		SimpleDateFormat dFormat = new SimpleDateFormat("d MMM yyyy")
		return dFormat.format(new Date())
	}
}
